"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.parse = void 0;
const recursive_descent_parser_js_1 = require("../../deps/deno.land/x/pbkit@v0.0.39/core/parser/recursive-descent-parser.js");
function parse(text) {
    const parser = (0, recursive_descent_parser_js_1.createRecursiveDescentParser)(text);
    const rules = acceptStatements(parser, [acceptRule]);
    return { ast: rules, parser };
}
exports.parse = parse;
function acceptStatement(parser, acceptStatementFn) {
    const { commentGroups: leadingComments } = skipWsAndSweepComments(parser);
    const statement = acceptStatementFn(parser, leadingComments);
    return statement;
}
function acceptStatements(parser, acceptStatementFns) {
    const statements = [];
    statements: while (true) {
        const { commentGroups: leadingComments } = skipWsAndSweepComments(parser);
        for (const acceptStatementFn of acceptStatementFns) {
            const statement = acceptStatementFn(parser, leadingComments);
            if (statement) {
                statements.push(statement);
                continue statements;
            }
        }
        break;
    }
    return statements;
}
function acceptRule(parser, leadingComments) {
    const head = acceptStatement(parser, acceptHead);
    if (!head)
        return;
    skipWsAndComments(parser);
    const tail = acceptStatement(parser, acceptTail);
    if (!tail)
        return;
    const trailingComments = acceptTrailingComments(parser);
    return {
        ...mergeSpans([
            head,
            tail,
            leadingComments,
            trailingComments,
        ]),
        head,
        tail,
        leadingComments,
        trailingComments,
    };
}
// @TODO: https://datatracker.ietf.org/doc/html/rfc3986#section-3
function acceptHead(parser, leadingComments) {
    const schemeOrAuthority = acceptSchemeOrAuthority(parser);
    if (schemeOrAuthority) {
        const { scheme, authority } = schemeOrAuthority;
        const path = acceptPath(parser);
        const trailingComments = acceptTrailingComments(parser);
        return {
            ...mergeSpans([
                scheme,
                authority,
                path,
                leadingComments,
            ]),
            scheme,
            authority,
            path,
            leadingComments,
            trailingComments,
        };
    }
    else {
        const path = acceptPath(parser);
        const trailingComments = acceptTrailingComments(parser);
        if (!path)
            return;
        return {
            ...mergeSpans([
                path,
                leadingComments,
            ]),
            path,
            leadingComments,
            trailingComments,
        };
    }
    function acceptSchemeOrAuthority(parser) {
        const loc = parser.loc;
        return acceptPatternAndThen(schemePattern, (token) => {
            if (!token)
                return;
            const isSchemeExist = parser.accept(":");
            if (!isSchemeExist)
                parser.loc = loc;
            return isSchemeExist
                ? { scheme: token, authority: acceptAuthority(parser) }
                : { authority: acceptAuthority(parser) };
        })(parser);
    }
}
function acceptAuthority(parser) {
    const slash = parser.accept(/^\/\//);
    const userinfoAndHost = acceptUserinfoAndHost(parser);
    if (!userinfoAndHost)
        return;
    const { userinfo, host } = userinfoAndHost;
    const port = parser.accept(":") && parser.accept(identPattern);
    return {
        ...mergeSpans([slash, userinfo, host, port]),
        userinfo,
        host,
        port,
    };
    function acceptUserinfoAndHost(parser) {
        return acceptPatternAndThen(identPattern, (token) => {
            if (!token)
                return;
            const isUserinfoExist = parser.accept("@");
            return isUserinfoExist
                ? { userinfo: token, host: parser.expect(identPattern) }
                : { host: token };
        })(parser);
    }
}
function acceptPath(parser) {
    const fragments = [];
    while (parser.accept("/")) {
        const fragment = acceptPathFragment(parser);
        if (fragment)
            fragments.push(fragment);
    }
    if (fragments.length === 0)
        return;
    return {
        ...mergeSpans(fragments),
        fragments,
    };
}
// @TODO: https://datatracker.ietf.org/doc/html/rfc3986#section-3.3
function acceptPathFragment(parser) {
    const isParamPathFragment = Boolean(parser.accept("["));
    const path = parser.accept(identPattern);
    isParamPathFragment && parser.expect("]");
    if (!path)
        return;
    return { type: isParamPathFragment ? "param" : "static", ...path };
}
function acceptTail(parser, leadingComments) {
    const openParens = parser.expect("{");
    const rules = acceptTailRules(parser);
    const closeParens = parser.expect("}");
    const trailingComments = acceptTrailingComments(parser);
    return {
        ...mergeSpans([
            openParens,
            closeParens,
            leadingComments,
            trailingComments,
        ]),
        openParens,
        rules,
        closeParens,
        leadingComments,
        trailingComments,
    };
}
function acceptTailRules(parser) {
    const tailRules = [];
    while (true) {
        const tailRule = acceptStatement(parser, acceptTailRule);
        if (!tailRule)
            break;
        tailRules.push(tailRule);
        skipWsAndComments(parser);
    }
    return tailRules;
}
function acceptTailRule(parser, leadingComments) {
    const tailType = parser.accept(/^(\?|#)/);
    if (!tailType)
        return;
    skipWsAndComments(parser);
    const matchType = parser.accept(/^(form|match)/);
    if (!matchType)
        return;
    skipWsAndComments(parser);
    switch (matchType.text) {
        case "form": {
            const openParens = parser.accept("{");
            // skipWsAndComments(parser);
            const pattern = acceptTailRuleFormPatternRules(parser);
            skipWsAndComments(parser);
            const closeParens = parser.accept("}");
            return {
                openParens,
                closeParens,
                tailType,
                matchType,
                pattern,
                leadingComments,
                trailingComments: acceptTrailingComments(parser),
            }; // @FIXME: typing!
        }
        case "match": {
            return {
                tailType,
                matchType,
                pattern: acceptTailRuleMatchPatternRules(parser),
                leadingComments,
                trailingComments: acceptTrailingComments(parser),
            };
        }
    }
}
function acceptTailRuleFormPatternRules(parser) {
    const rules = [];
    while (!parser.try(/^\s*}/)) {
        const { commentGroups: leadingComments } = skipWsAndSweepComments(parser);
        const key = choice([acceptIdent, acceptStrLit])(parser);
        if (!key)
            break;
        skipWsAndComments(parser);
        const array = parser.accept("[]");
        skipWsAndComments(parser);
        const equal = parser.accept("=");
        if (equal) {
            const values = acceptTailRuleMatchPatternRules(parser);
            const trailingComments = acceptTrailingComments(parser);
            rules.push({
                ...mergeSpans([
                    key,
                    ...values,
                    leadingComments,
                    trailingComments,
                ]),
                key,
                array,
                value: values,
                leadingComments,
                trailingComments,
            });
        }
        else {
            const trailingComments = acceptTrailingComments(parser);
            rules.push({
                ...mergeSpans([
                    key,
                    leadingComments,
                    trailingComments,
                ]),
                key,
                array,
                leadingComments,
                trailingComments,
            });
        }
    }
    return rules;
}
function acceptTailRuleMatchPatternRules(parser) {
    let loc = parser.loc;
    const rules = [];
    parser.accept("|");
    do {
        const { commentGroups: leadingComments } = skipWsAndSweepComments(parser);
        const rule = choice([acceptRegex, acceptIdent, acceptStrLit])(parser);
        rules.push({
            ...rule,
            leadingComments,
            trailingComments: acceptTrailingComments(parser),
        });
        loc = parser.loc;
        skipWsAndComments(parser);
    } while (parser.accept("|"));
    parser.loc = loc;
    return rules;
}
function acceptIdent(parser) {
    const ident = parser.accept(identPattern);
    if (!ident)
        return;
    return { ...ident, type: "id" };
}
function acceptStrLit(parser) {
    const strLit = parser.accept(strLitPattern);
    if (!strLit)
        return;
    return { ...strLit, type: "string" };
}
function acceptRegex(parser) {
    const regex = parser.accept(regexPattern);
    if (!regex)
        return;
    return { ...regex, type: "regex" };
}
function mergeSpans(spans) {
    let start = Infinity;
    let end = -Infinity;
    for (let i = 0; i < spans.length; ++i) {
        if (spans[i] == null)
            continue;
        const span = Array.isArray(spans[i])
            ? mergeSpans(spans[i])
            : spans[i];
        start = Math.min(start, span.start);
        end = Math.max(end, span.end);
    }
    return { start, end };
}
function acceptPatternAndThen(pattern, then) {
    return function accept(parser) {
        const token = parser.accept(pattern);
        return then(token);
    };
}
function choice(acceptFns) {
    return function accept(parser) {
        for (const acceptFn of acceptFns) {
            const node = acceptFn(parser);
            if (node)
                return node;
        }
    };
}
function many(parser, acceptFn) {
    const nodes = [];
    let node;
    while (node = acceptFn(parser))
        nodes.push(node);
    return nodes;
}
const whitespacePattern = /^\s+/;
const whitespaceWithoutNewlinePattern = /^[ \f\t\v\u00a0\u1680\u2000-\u200a\u2028\u2029\u202f\u205f\u3000\ufeff]+/;
const newlinePattern = /^\r?\n/;
const identPattern = /^(?:[a-zA-Z0-9\-._~]|%[0-9a-fA-F]{2})+/;
const strLitPattern = /^'.*?'/;
const multilineCommentPattern = /^\/\*(?:.|\r?\n)*?\*\//;
const documentCommentPattern = /^\/\*\*(?:.|\r?\n\s*\*)*(\r?\n\s*)\*\//;
const singlelineCommentPattern = /^\/\/.*(?:\r?\n|$)/;
const schemePattern = /^[a-z][a-z0-9\+-\.]*/i;
const regexPattern = /^\/(?:(?![*+?])(?:[^\r\n\[/\\]|\\.|\[(?:[^\r\n\]\\]|\\.)*\])+)\/[a-z]*/;
function acceptSpecialToken(parser, type, pattern = identPattern) {
    const token = parser.accept(pattern);
    if (!token)
        return;
    return { type, ...token };
}
function acceptCommentGroup(parser) {
    const loc = parser.loc;
    const comments = [];
    while (true) {
        const whitespace = parser.accept(whitespaceWithoutNewlinePattern);
        if (whitespace)
            continue;
        const documentComment = acceptSpecialToken(parser, "document-comment", documentCommentPattern);
        if (documentComment) {
            comments.push(documentComment);
            continue;
        }
        const multilineComment = acceptSpecialToken(parser, "multiline-comment", multilineCommentPattern);
        if (multilineComment) {
            comments.push(multilineComment);
            continue;
        }
        const singlelineComment = acceptSpecialToken(parser, "singleline-comment", singlelineCommentPattern);
        if (singlelineComment) {
            comments.push(singlelineComment);
            continue;
        }
        break;
    }
    if (comments.length < 1) {
        parser.loc = loc;
        return;
    }
    return {
        ...mergeSpans(comments),
        type: "comment-group",
        comments,
    };
}
function acceptTrailingComments(parser) {
    const loc = parser.loc;
    const comments = [];
    while (true) {
        const whitespace = parser.accept(whitespaceWithoutNewlinePattern);
        if (whitespace)
            continue;
        const newline = parser.accept(newlinePattern);
        if (newline)
            break;
        const documentComment = acceptSpecialToken(parser, "document-comment", documentCommentPattern);
        if (documentComment) {
            comments.push(documentComment);
            continue;
        }
        const multilineComment = acceptSpecialToken(parser, "multiline-comment", multilineCommentPattern);
        if (multilineComment) {
            comments.push(multilineComment);
            continue;
        }
        const singlelineComment = acceptSpecialToken(parser, "singleline-comment", singlelineCommentPattern);
        if (singlelineComment) {
            comments.push(singlelineComment);
            break;
        }
        break;
    }
    if (comments.length < 1) {
        parser.loc = loc;
        return [];
    }
    return [{
            ...mergeSpans(comments),
            type: "comment-group",
            comments,
        }];
}
function skipWsAndSweepComments(parser) {
    const commentGroups = [];
    let trailingNewline = false;
    parser.accept(whitespacePattern);
    while (true) {
        const commentGroup = acceptCommentGroup(parser);
        if (commentGroup) {
            commentGroups.push(commentGroup);
            trailingNewline = false;
            continue;
        }
        const whitespace = parser.accept(whitespaceWithoutNewlinePattern);
        if (whitespace)
            continue;
        const newline = parser.accept(newlinePattern);
        if (newline) {
            trailingNewline = true;
            continue;
        }
        break;
    }
    return {
        commentGroups,
        trailingNewline,
    };
}
function skipWsAndComments(parser) {
    while (true) {
        const whitespace = parser.accept(whitespacePattern);
        if (whitespace)
            continue;
        const multilineComment = acceptSpecialToken(parser, "multiline-comment", multilineCommentPattern);
        if (multilineComment)
            continue;
        const singlelineComment = acceptSpecialToken(parser, "singleline-comment", singlelineCommentPattern);
        if (singlelineComment)
            continue;
        break;
    }
    return;
}
